
(*************************** BITOP.MOD ****************************)
IMPLEMENTATION MODULE BitOp;
(*
**  Filename:   BitOp.MOD (Modula-2)
**
**  Author:     Gregory S. Vigneault, z180@hotmail.com
**              1998.Dec.07
*)

IMPORT MathLib0;

CONST

  BYTEBIT = SIZE(CHAR) * 8 - 1;     (* high bit number in CHARs     *)
  WORDBIT = SIZE(CARDINAL) * 8 - 1; (* high bit number in CARDINALs *)

TYPE

  ULONG = RECORD
            CASE (*union*): CARDINAL OF
              0 : long    :LONGCARD;
            | 1 : LO, HI  :BITSET; (* used as little-endian words *)
            END;                   (* reverse HI,LO for big-endian*)
          END;

(*----------------------------------------------------------------*)

PROCEDURE ANDW (val, mask :CARDINAL) :CARDINAL;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN CARDINAL( BITSET(val) * BITSET(mask) );
  END ANDW;

(*----------------------------------------------------------------*)

PROCEDURE ORW (val, mask :CARDINAL) :CARDINAL;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN CARDINAL( BITSET(val) + BITSET(mask) );
  END ORW;

(*----------------------------------------------------------------*)

PROCEDURE XORW (val, mask :CARDINAL) :CARDINAL;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN CARDINAL( BITSET(val) / BITSET(mask) );
  END XORW;

(*----------------------------------------------------------------*)

PROCEDURE NOTW (val :CARDINAL) :CARDINAL;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN CARDINAL( BITSET(val) / {0..15} );
  END NOTW;

(*----------------------------------------------------------------*)

PROCEDURE IMPW (p,q :CARDINAL) :CARDINAL;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN NOTW( ANDW( p, NOTW(q) ));
  END IMPW;

(*----------------------------------------------------------------*)

PROCEDURE SHLW (val :CARDINAL; shifts :CARDINAL) :CARDINAL;
  BEGIN
    WHILE shifts # 0 DO
      (* CarryFlag := WORDBIT IN BITSET(val); *)
      INC(val,val);
      DEC(shifts);
    END;
    RETURN val;
  END SHLW;

(*----------------------------------------------------------------*)

PROCEDURE SHRW (val :CARDINAL; shifts :CARDINAL) :CARDINAL;
  BEGIN
    WHILE shifts # 0 DO
      (* CarryFlag := 0 IN BITSET(val); *)
      val := val DIV 2;
      DEC(shifts);
    END;
    RETURN val;
  END SHRW;

(*================================================================*)

PROCEDURE ANDB (val, mask :CHAR) :CHAR;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN CHR(CARDINAL(BITSET(ORD(val)) * BITSET(ORD(mask))));
  END ANDB;

(*----------------------------------------------------------------*)

PROCEDURE ORB (val, mask :CHAR) :CHAR;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN CHR(CARDINAL(BITSET(ORD(val)) + BITSET(ORD(mask))));
  END ORB;

(*----------------------------------------------------------------*)

PROCEDURE XORB (val, mask :CHAR) :CHAR;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN CHR(CARDINAL(BITSET(ORD(val)) / BITSET(ORD(mask))));
  END XORB;

(*----------------------------------------------------------------*)

PROCEDURE NOTB (val :CHAR) :CHAR;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN CHR(CARDINAL(BITSET(ORD(val)) / {0..7}));
  END NOTB;

(*----------------------------------------------------------------*)

PROCEDURE IMPB (p,q :CHAR) :CHAR;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN NOTB( ANDB( p, NOTB(q) ));
  END IMPB;

(*----------------------------------------------------------------*)

PROCEDURE SHLB (val :CHAR; shifts :CARDINAL) :CHAR;
  BEGIN
    WHILE shifts # 0 DO
      (* CarryFlag := BYTEBIT IN BITSET(ORD(val)); *)
      val := CHR(ORD(val) + ORD(val));
      DEC(shifts);
    END;
    RETURN val;
  END SHLB;

(*----------------------------------------------------------------*)

PROCEDURE SHRB (val :CHAR; shifts :CARDINAL) :CHAR;
  BEGIN
    WHILE shifts # 0 DO
      (* CarryFlag := 0 IN BITSET(ORD(val)); *)
      val := CHR(ORD(val) DIV 2);
      DEC(shifts);
    END;
    RETURN val;
  END SHRB;

(*================================================================*)

PROCEDURE ANDL (val, mask :LONGCARD) :LONGCARD;
  VAR u1, u2 :ULONG;
  BEGIN
    (* CarryFlag := FALSE; *)
    u1.long := val;
    u2.long := mask;
    u1.LO := u1.LO * u2.LO;
    u1.HI := u1.HI * u2.HI;
    RETURN u1.long;
  END ANDL;

(*----------------------------------------------------------------*)

PROCEDURE ORL (val, mask :LONGCARD) :LONGCARD;
  VAR u1, u2 :ULONG;
  BEGIN
    (* CarryFlag := FALSE; *)
    u1.long := val;
    u2.long := mask;
    u1.LO := u1.LO + u2.LO;
    u1.HI := u1.HI + u2.HI;
    RETURN u1.long;
  END ORL;

(*----------------------------------------------------------------*)

PROCEDURE XORL (val, mask :LONGCARD) :LONGCARD;
  VAR u1, u2 :ULONG;
  BEGIN
    (* CarryFlag := FALSE; *)
    u1.long := val;
    u2.long := mask;
    u1.LO := u1.LO / u2.LO;
    u1.HI := u1.HI / u2.HI;
    RETURN u1.long;
  END XORL;

(*----------------------------------------------------------------*)

PROCEDURE NOTL (val :LONGCARD) :LONGCARD;
  VAR u :ULONG;
  BEGIN
    (* CarryFlag := FALSE; *)
    u.long := val;
    u.LO := u.LO / {0..15};
    u.HI := u.HI / {0..15};
    RETURN u.long;
  END NOTL;

(*----------------------------------------------------------------*)

PROCEDURE IMPL (p,q :LONGCARD) :LONGCARD;
  BEGIN
    (* CarryFlag := FALSE; *)
    RETURN NOTL( ANDL( p, NOTL(q) ));
  END IMPL;

(*----------------------------------------------------------------*)

PROCEDURE SHLL (val :LONGCARD; shifts :CARDINAL) :LONGCARD;
  VAR u :ULONG;
  BEGIN
    u.long := val;
    WHILE shifts # 0 DO
      (* CarryFlag := WORDBIT IN u.HI; *)
      u.long := u.long + u.long;
      DEC(shifts);
    END;
    RETURN u.long;
  END SHLL;

(*----------------------------------------------------------------*)

PROCEDURE SHRL (val :LONGCARD; shifts :CARDINAL) :LONGCARD;
  VAR u :ULONG;
  BEGIN
    u.long := val;
    WHILE shifts # 0 DO
      (* CarryFlag := 0 IN u.LO; *)
      u.long := u.long DIV 2L;
      DEC(shifts);
    END;
    RETURN u.long;
  END SHRL;

(*----------------------------------------------------------------*)

BEGIN

END BitOp.
(*************************** BITOP.MOD ****************************)

