(************************** ISAAC.MOD *****************************)
IMPLEMENTATION MODULE ISAAC;
(*
** Filename: ISAAC.MOD (Modula-2)
**
** Purpose: Random number generator ISAAC (Indirection, Shift,
** Accumulate, Add, and Count), as developed by Bob Jenkins.
**
** Gregory S. Vigneault, z180@hotmail.com
*)
FROM BitOp  IMPORT ANDL, ANDW, SHLL, SHRL, XORL;

(*----------------------------------------------------------------*)

PROCEDURE isaac (ctx :randctxPTR);
  CONST INDC = LONG(4 * (RANDSIZ-1));
  VAR a,b,x,y :LONGCARD;
      m,mm,m2,r,mend :CARDINAL;

      PROCEDURE ind (x :LONGCARD) :LONGCARD;
        BEGIN
          RETURN ctx^.randmem[mm + SHORT(ANDL(x, INDC)) DIV 4];
        END ind;

      PROCEDURE rngstep (mix :LONGCARD);
        BEGIN
          x := ctx^.randmem[m];
          a := XORL(a,mix) + ctx^.randmem[m2];
          INC(m2);
          y := a + b + ind(x);
          ctx^.randmem[m] := y;
          INC(m);
          b := x + ind( SHRL(y,RANDSIZL) );
          ctx^.randrsl[r] := b;
          INC(r);
        END rngstep;

  BEGIN (*isaac*)

    a := ctx^.randa;
    INC(ctx^.randc);
    b := ctx^.randb + ctx^.randc;

    mm := 0;  (* index for ctx^.randmem[] *)
    r := 0;   (* index for ctx^.randrsl[] *)

    m := mm;
    m2 := m + RANDSIZ DIV 2;
    mend := m2;

    WHILE m < mend DO
      rngstep( SHLL(a,13) );
      rngstep( SHRL(a, 6) );
      rngstep( SHLL(a, 2) );
      rngstep( SHRL(a,16) );
    END;

    m2 := mm;
    WHILE m2 < mend DO
      rngstep( SHLL(a,13) );
      rngstep( SHRL(a, 6) );
      rngstep( SHLL(a, 2) );
      rngstep( SHRL(a,16) );
    END;

   ctx^.randb := b;
   ctx^.randa := a;

  END isaac;

(*----------------------------------------------------------------*)

PROCEDURE randinit (ctx :randctxPTR; flag :BOOLEAN);
(* IF flag=TRUE then use contents of randrsl[] to initialize mm[] *)
  VAR a,b,c,d,e,f,g,h :LONGCARD;
      i :CARDINAL;

      PROCEDURE mix;
        BEGIN
          a := XORL(a, SHLL(b,11));   d := d + a;   b := b + c;
          b := XORL(b, SHRL(c, 2));   e := e + b;   c := c + d;
          c := XORL(c, SHLL(d, 8));   f := f + c;   d := d + e;
          d := XORL(d, SHRL(e,16));   g := g + d;   e := e + f;
          e := XORL(e, SHLL(f,10));   h := h + e;   f := f + g;
          f := XORL(f, SHRL(g, 4));   a := a + f;   g := g + h;
          g := XORL(g, SHLL(h, 8));   b := b + g;   h := h + a;
          h := XORL(h, SHRL(a, 9));   c := c + h;   a := a + b;
        END mix;

  BEGIN (*randinit*)

    ctx^.randc := 0L;
    ctx^.randb := 0L;
    ctx^.randa := 0L;

    a := 2654435769;            (* 9E3779B9H: The Golden Ratio *)
    b := a;  c := a;  d := a;
    e := a;  f := a;  g := a;
    h := a;

    FOR i := 0 TO 3 DO  (* scramble it *)
      mix;
    END;

    IF flag THEN  (* init using contents of ctx^.randrsl[] as seed *)

      FOR i := 0 TO RANDSIZ-1 BY 8 DO
        a := a + ctx^.randrsl[i  ];
        b := b + ctx^.randrsl[i+1];
        c := c + ctx^.randrsl[i+2];
        d := d + ctx^.randrsl[i+3];
        e := e + ctx^.randrsl[i+4];
        f := f + ctx^.randrsl[i+5];
        g := g + ctx^.randrsl[i+6];
        h := h + ctx^.randrsl[i+7];
        mix;
        ctx^.randmem[i  ] := a;
        ctx^.randmem[i+1] := b;
        ctx^.randmem[i+2] := c;
        ctx^.randmem[i+3] := d;
        ctx^.randmem[i+4] := e;
        ctx^.randmem[i+5] := f;
        ctx^.randmem[i+6] := g;
        ctx^.randmem[i+7] := h;
      END;

      (* 2nd pass makes all of seed affect all of ctx^.randmem[] *)

      FOR i := 0 TO RANDSIZ-1 BY 8 DO
        a := a + ctx^.randmem[i  ];
        b := b + ctx^.randmem[i+1];
        c := c + ctx^.randmem[i+2];
        d := d + ctx^.randmem[i+3];
        e := e + ctx^.randmem[i+4];
        f := f + ctx^.randmem[i+5];
        g := g + ctx^.randmem[i+6];
        h := h + ctx^.randmem[i+7];
        mix;
        ctx^.randmem[i  ] := a;
        ctx^.randmem[i+1] := b;
        ctx^.randmem[i+2] := c;
        ctx^.randmem[i+3] := d;
        ctx^.randmem[i+4] := e;
        ctx^.randmem[i+5] := f;
        ctx^.randmem[i+6] := g;
        ctx^.randmem[i+7] := h;
      END;

    ELSE

      (* fill in messy stuff *)

      FOR i := 0 TO RANDSIZ-1 BY 8 DO
        mix;
        ctx^.randmem[i  ] := a;
        ctx^.randmem[i+1] := b;
        ctx^.randmem[i+2] := c;
        ctx^.randmem[i+3] := d;
        ctx^.randmem[i+4] := e;
        ctx^.randmem[i+5] := f;
        ctx^.randmem[i+6] := g;
        ctx^.randmem[i+7] := h;
      END;

    END; (*IF flag*)

    isaac(ctx);   (* fill in the first set of results *)

    ctx^.randcnt := LONG(RANDSIZ); (* prep to use 1st set of results *)

  END randinit;

(*----------------------------------------------------------------*)

PROCEDURE rand (r :randctxPTR) :LONGCARD;
  (* Call rand() to retrieve a single 32-bit random value. *)
  BEGIN

    IF r^.randcnt = 0L THEN
      isaac(r);
      r^.randcnt := LONG(RANDSIZ - 1);
    ELSE
      DEC(r^.randcnt);
    END;

    RETURN r^.randrsl[SHORT(r^.randcnt)];

  END rand;

(*================================================================*)

BEGIN (*MODULE Rand*)

  ;

END ISAAC.
(************************** ISAAC.MOD *****************************)


